/*
 * Copyright (C) 2005-2014 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file crypt_common.c Crypto functions independent of openssl/libgcrypt. */

#include <regex.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "crypt.h"
#include "crypt_backend.h"

/** If the key begins with this text, we treat it as an ssh key. */
#define KEY_TYPE_TXT "ssh-rsa"

/**
 * Check if given buffer starts with a ssh rsa key signature.
 *
 * \param data The buffer.
 * \param size Number of data bytes.
 *
 * \return Number of header bytes to be skipped on success, zero if
 * ssh rsa signature was not found.
 */
size_t is_ssh_rsa_key(char *data, size_t size)
{
	char *cp;

	if (size < strlen(KEY_TYPE_TXT) + 2)
		return 0;
	cp = memchr(data, ' ', size);
	if (cp == NULL)
		return 0;
	if (strncmp(KEY_TYPE_TXT, data, strlen(KEY_TYPE_TXT)))
		return 0;
	cp++;
	if (cp >= data + size)
		return 0;
	if (*cp == '\0')
		return 0;
	return cp - data;
}

/*
 * This base64/uudecode stuff below is taken from openssh-5.2p1, Copyright (c)
 * 1996 by Internet Software Consortium.  Portions Copyright (c) 1995 by
 * International Business Machines, Inc.
 */

static const char Base64[] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static const char Pad64 = '=';

/**
 * base64-decode a buffer.
 *
 * \param src The buffer to decode.
 * \param target Result is stored here.
 * \param targsize Number of bytes of \a target.
 *
 * Skips all whitespace anywhere. Converts characters, four at a time, starting
 * at (or after) src from base - 64 numbers into three 8 bit bytes in the
 * target area.
 *
 * \return The number of data bytes stored at the target, -E_BASE64 on errors.
 */
int base64_decode(char const *src, unsigned char *target, size_t targsize)
{
	unsigned int tarindex, state;
	int ch;
	char *pos;

	state = 0;
	tarindex = 0;

	while ((ch = *src++) != '\0') {
		if (para_isspace(ch)) /* Skip whitespace anywhere. */
			continue;

		if (ch == Pad64)
			break;

		pos = strchr(Base64, ch);
		if (pos == NULL) /* A non-base64 character. */
			return -E_BASE64;

		switch (state) {
		case 0:
			if (target) {
				if (tarindex >= targsize)
					return -E_BASE64;
				target[tarindex] = (pos - Base64) << 2;
			}
			state = 1;
			break;
		case 1:
			if (target) {
				if (tarindex + 1 >= targsize)
					return -E_BASE64;
				target[tarindex]   |=  (pos - Base64) >> 4;
				target[tarindex+1]  = ((pos - Base64) & 0x0f)
							<< 4 ;
			}
			tarindex++;
			state = 2;
			break;
		case 2:
			if (target) {
				if (tarindex + 1 >= targsize)
					return -E_BASE64;
				target[tarindex]   |=  (pos - Base64) >> 2;
				target[tarindex+1]  = ((pos - Base64) & 0x03)
							<< 6;
			}
			tarindex++;
			state = 3;
			break;
		case 3:
			if (target) {
				if (tarindex >= targsize)
					return -E_BASE64;
				target[tarindex] |= (pos - Base64);
			}
			tarindex++;
			state = 0;
			break;
		}
	}

	/*
	 * We are done decoding Base-64 chars.  Let's see if we ended
	 * on a byte boundary, and/or with erroneous trailing characters.
	 */

	if (ch == Pad64) {		/* We got a pad char. */
		ch = *src++;		/* Skip it, get next. */
		switch (state) {
		case 0:		/* Invalid = in first position */
		case 1:		/* Invalid = in second position */
			return -E_BASE64;

		case 2:		/* Valid, means one byte of info */
			/* Skip any number of spaces. */
			for (; ch != '\0'; ch = *src++)
				if (!isspace(ch))
					break;
			/* Make sure there is another trailing = sign. */
			if (ch != Pad64)
				return -E_BASE64;
			ch = *src++;		/* Skip the = */
			/* Fall through to "single trailing =" case. */
			/* FALLTHROUGH */

		case 3:		/* Valid, means two bytes of info */
			/*
			 * We know this char is an =.  Is there anything but
			 * whitespace after it?
			 */
			for (; ch != '\0'; ch = *src++)
				if (!isspace(ch))
					return -E_BASE64;

			/*
			 * Now make sure for cases 2 and 3 that the "extra"
			 * bits that slopped past the last full byte were
			 * zeros.  If we don't check them, they become a
			 * subliminal channel.
			 */
			if (target && target[tarindex] != 0)
				return -E_BASE64;
		}
	} else {
		/*
		 * We ended by seeing the end of the string.  Make sure we
		 * have no partial bytes lying around.
		 */
		if (state != 0)
			return -E_BASE64;
	}

	return tarindex;
}

/**
 * uudecode a buffer.
 *
 * \param src The buffer to decode.
 * \param target Result buffer.
 * \param targsize The length of \a target in bytes.
 *
 * This is just a simple wrapper for base64_decode() which strips whitespace.
 *
 * \return The return value of the underlying call to base64_decode().
 */
int uudecode(const char *src, unsigned char *target, size_t targsize)
{
	int len;
	char *encoded, *p;

	/* copy the 'readonly' source */
	encoded = para_strdup(src);
	/* skip whitespace and data */
	for (p = encoded; *p == ' ' || *p == '\t'; p++)
		;
	for (; *p != '\0' && *p != ' ' && *p != '\t'; p++)
		;
	/* and remove trailing whitespace because base64_decode needs this */
	*p = '\0';
	len = base64_decode(encoded, target, targsize);
	free(encoded);
	return len;
}

/**
 * Read a 4-byte number from a buffer in big-endian format.
 *
 * \param vp The buffer.
 *
 * The byte-order of the buffer is expected to be big-endian, unlike read_u32()
 * of portable_io.h which expects little endian.
 *
 * \return The 32 bit number given by \a vp.
 */
uint32_t read_ssh_u32(const void *vp)
{
	const unsigned char *p = (const unsigned char *)vp;
	uint32_t v;

	v  = (uint32_t)p[0] << 24;
	v |= (uint32_t)p[1] << 16;
	v |= (uint32_t)p[2] << 8;
	v |= (uint32_t)p[3];

	return v;
}

/**
 * Sanity checks for the header of an ssh key.
 *
 * \param blob The buffer.
 * \param blen The number of bytes of \a blob.
 *
 * This performs some checks to make sure we really have an ssh key. It also
 * computes the offset in bytes of the start of the key values (modulus,
 * exponent..).
 *
 * \return The number of bytes to skip until the start of the first encoded
 * number (usually 11).
 */
int check_ssh_key_header(const unsigned char *blob, int blen)
{
	const unsigned char *p = blob, *end = blob + blen;
	uint32_t rlen;

	if (p + 4 > end)
		return -E_SSH_KEY_HEADER;
	rlen = read_ssh_u32(p);
	p += 4;
	if (p + rlen < p)
		return -E_SSH_KEY_HEADER;
	if (p + rlen > end)
		return -E_SSH_KEY_HEADER;
	if (rlen < strlen(KEY_TYPE_TXT))
		return -E_SSH_KEY_HEADER;
	PARA_DEBUG_LOG("type: %s, rlen: %d\n", p, rlen);
	if (strncmp((char *)p, KEY_TYPE_TXT, strlen(KEY_TYPE_TXT)))
		return -E_SSH_KEY_HEADER;
	return 4 + rlen;
}

/**
 * Check existence and permissions of a key file.
 *
 * \param file The path of the key file.
 * \param private_key Whether this is a private key.
 *
 * This checks whether the file exists. If it is a private key, we additionally
 * check that the permissions are restrictive enough. It is considered an error
 * if we own the file and it is readable for others.
 *
 * \return Standard.
 */
int check_key_file(const char *file, bool private_key)
{
	struct stat st;

	if (stat(file, &st) != 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	if (!private_key)
		return 0;
	if ((st.st_uid == getuid()) && (st.st_mode & 077) != 0)
		return -E_KEY_PERM;
	return 1;
}

void hash_to_asc(unsigned char *hash, char *asc)
{
	int i;
	const char hexchar[] = "0123456789abcdef";

	for (i = 0; i < HASH_SIZE; i++) {
		asc[2 * i] = hexchar[hash[i] >> 4];
		asc[2 * i + 1] = hexchar[hash[i] & 0xf];
	}
	asc[2 * HASH_SIZE] = '\0';
}

int hash_compare(unsigned char *h1, unsigned char *h2)
{
	int i;

	for (i = 0; i < HASH_SIZE; i++) {
		if (h1[i] < h2[i])
			return -1;
		if (h1[i] > h2[i])
			return 1;
	}
	return 0;
}
