/*
 * Copyright (C) 2005-2014 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file http_recv.c paraslash's http receiver */

#include <regex.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>

#include "para.h"
#include "error.h"
#include "http.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "recv.h"
#include "http_recv.cmdline.h"
#include "net.h"
#include "string.h"
#include "fd.h"

/**
 * the possible states of a http receiver node
 *
 * \sa receiver_node
 */
enum http_recv_status {HTTP_CONNECTED, HTTP_SENT_GET_REQUEST, HTTP_STREAMING};

/**
 * Data specific to the http receiver.
 *
 * Each running instance of the http receiver reserves space for one such struct.
 */
struct private_http_recv_data {
	/**
	 * The current status of the http receiver node.
	 *
	 * It gets initialized to \p HTTP_CONNECTED by the open function of the
	 * http receiver.
	 *
	 * \sa receiver::open, receiver_node.
	 */
	enum http_recv_status status;
};

static char *make_request_msg(void)
{
	char *ret, *hn = para_hostname();
	ret = make_message("%s1.0\nHost: %s\nUser-Agent: para_recv/%s\n\n\n",
		HTTP_GET_MSG, hn, PACKAGE_VERSION);
	free(hn);
	return ret;
}

static void http_recv_pre_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);
	struct private_http_recv_data *phd = rn->private_data;

	if (generic_recv_pre_select(s, t) <= 0)
		return;
	if  (phd->status == HTTP_CONNECTED)
		para_fd_set(rn->fd, &s->wfds, &s->max_fileno);
	else
		para_fd_set(rn->fd, &s->rfds, &s->max_fileno);
}

/*
 * Establish the http connection. If already established, fill the buffer pool
 * area with data read from the socket. In any case, update the state of the
 * connection if necessary.
 */
static int http_recv_post_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);
	struct private_http_recv_data *phd = rn->private_data;
	struct btr_node *btrn = rn->btrn;
	int ret, iovcnt;
	struct iovec iov[2];
	size_t num_bytes;

	ret = task_get_notification(t);
	if (ret < 0)
		goto out;
	ret = btr_node_status(btrn, 0, BTR_NT_ROOT);
	if (ret < 0)
		goto out;
	if (ret == 0)
		return 0;
	if (phd->status == HTTP_CONNECTED) {
		char *rq;
		if (!FD_ISSET(rn->fd, &s->wfds))
			return 0;
		rq = make_request_msg();
		PARA_INFO_LOG("sending http request\n");
		ret = write_va_buffer(rn->fd, "%s", rq);
		free(rq);
		if (ret < 0)
			goto out;
		phd->status = HTTP_SENT_GET_REQUEST;
		return 0;
	}
	if (phd->status == HTTP_SENT_GET_REQUEST) {
		ret = read_pattern(rn->fd, HTTP_OK_MSG, strlen(HTTP_OK_MSG), &s->rfds);
		if (ret < 0)
			goto out;
		if (ret == 0)
			return 0;
		PARA_INFO_LOG("received ok msg, streaming\n");
		phd->status = HTTP_STREAMING;
		return 0;
	}
	ret = -E_HTTP_RECV_OVERRUN;
	iovcnt = btr_pool_get_buffers(rn->btrp, iov);
	if (iovcnt == 0)
		goto out;
	ret = readv_nonblock(rn->fd, iov, iovcnt, &s->rfds, &num_bytes);
	if (num_bytes == 0)
		goto out;
	if (num_bytes <= iov[0].iov_len) /* only the first buffer was filled */
		btr_add_output_pool(rn->btrp, num_bytes, btrn);
	else { /* both buffers contain data */
		btr_add_output_pool(rn->btrp, iov[0].iov_len, btrn);
		btr_add_output_pool(rn->btrp, num_bytes - iov[0].iov_len, btrn);
	}
out:
	if (ret < 0)
		btr_remove_node(&rn->btrn);
	return ret;
}

static void http_recv_close(struct receiver_node *rn)
{
	close(rn->fd);
	btr_pool_free(rn->btrp);
	free(rn->private_data);
}

static void *http_recv_parse_config(int argc, char **argv)
{
	struct http_recv_args_info *tmp = para_calloc(sizeof(*tmp));

	http_recv_cmdline_parser(argc, argv, tmp);
	return tmp;
}

static int http_recv_open(struct receiver_node *rn)
{
	struct private_http_recv_data *phd;
	struct http_recv_args_info *conf = rn->conf;
	int fd, ret = para_connect_simple(IPPROTO_TCP, conf->host_arg,
						       conf->port_arg);

	if (ret < 0)
		return ret;
	fd = ret;
	ret = mark_fd_nonblocking(fd);
	if (ret < 0) {
		close(fd);
		return ret;
	}
	rn->private_data = phd = para_calloc(sizeof(struct private_http_recv_data));
	rn->fd = fd;
	phd->status = HTTP_CONNECTED;
	rn->btrp = btr_pool_new("http_recv", 320 * 1024);
	return 1;
}

static void http_recv_free_config(void *conf)
{
	http_recv_cmdline_parser_free(conf);
	free(conf);
}

/**
 * The init function of the http receiver.
 *
 * \param r Pointer to the receiver struct to initialize.
 *
 * This initializes all function pointers of \a r.
 */
void http_recv_init(struct receiver *r)
{
	struct http_recv_args_info dummy;

	http_recv_cmdline_parser_init(&dummy);
	r->open = http_recv_open;
	r->close = http_recv_close;
	r->pre_select = http_recv_pre_select;
	r->post_select = http_recv_post_select;
	r->parse_config = http_recv_parse_config;
	r->free_config = http_recv_free_config;
	r->help = (struct ggo_help)DEFINE_GGO_HELP(http_recv);
	http_recv_cmdline_parser_free(&dummy);
}
